@@ -1,4 +1,3 @@
-#
 #    Do not delete!
 #  File name   	HotDeckImputation.R
 #  Part of:		   HotDeckImputation (GNU R contributed package)
@@ -6,22 +5,25 @@
 #  Copyright:		Dieter William Joenssen
 #  Email:			Dieter.Joenssen@TU-Ilmenau.de
 #  Created:		   14 May 2013
-#  Last Update: 	11 August 2013
+#  Last Update: 	31 August 2014
 #  Description:	R code for Package HotDeckImputation. Implemented functionions include following:
 #                 Topics:
 #                 -impute.NN_HD           ~ An extremely comprehensive implementation Nearest-Neighbor hot deck algorithms
 #                 -impute.mean            ~ Imputes the mean value of the complete cases in a variable for the missing values
 #                 -reweight.data          ~ Reweights data with given wieghts, preprocessing for distance calculation
 #				  -match.d_r_vam          ~ performs the Vogel's approximation method matching for donors-recipients
+#              -match.d_r_odd          ~ performs the ODD matching for donors-recipients
+#                 -impute.cps_HD          ~ performs the simple cps sequential hot deck method
+#                 -impute.fast_NN_HD      ~ performs fast version of nearest neighbor hot deck, not precomputing the distance matrix
 
 
 ##Description of impute.NN_HD
 {
-# data is data, requires numeric matrix
+# DATA is data, requires numeric matrix
 # distance is either:
 #     numeric matrix       ---   predefined donors x recipients distance matrix
 #                                (row/colnames must correspond to object number in data)
-#     string length=1      ---   defining dinstance metric to be used ("man","eukl","tscheb","mahal")
+#     string length=1      ---   defining distance metric to be used ("man","eukl","tscheb","mahal")
 #     numeric length=1       ---   defining minkovski parameter
 # weights is either:
 #     string length=1        ---   defining reweighting type ("range", "sd", "var", "none")
@@ -33,30 +35,31 @@
 #     string length=1     ---   defining compensation of missing values for distance calculation
 #                                ("rw_dist","mean","rseq","rsim")
 # donor_limit is:
-#     numeric length=1      ---   Interpretaion of value is range dependent:
+#     numeric length=1      ---   Interpretation of value is range dependent:
 #                                - if value is in (0,1) then it is interpreted as a dynamic donor limit 
 #                                  i.e. .5 means any 1 donor may serve 50% of all recipients
 #                                  in case of fractional results, this is rounded up
 #                                - if value is 1 or larger then it is interpreted as a static donor limit
 #                                  i.e. 2 means any 1 donor may serve up to 2 recipients
 #                                - Inf is unlimited donor usage
-#                                - Fractional parts for numers larger than 1 are discarded
+#                                - Fractional parts for numbers larger than 1 are discarded
 # optimal_donor:
 #  string length=1        ---   defines how the optimal donor is found when a donor limit is used: (problem explained extensively elsewhere)
 #                                - "no"      recipients are allocated their donor by order they apear in the data
 #                                - "rand"    recipients are allocated their donor by random order
-#                                - "mmin"    matrix minmum method is used to allocate donors to the recipients
+#                                - "mmin"    matrix minimum method is used to allocate donors to the recipients
 #                                - "vam"     Vogel's approximation method is used to allocate donors to the recipients
-#                                - "modi"    Modified Stepingstone algorithm is used
+#                                - "odd"     Optimal distribution of donors is used
+#                                - "modi"    Modified Steppingstone algorithm is used
 #                                ("no", "rand","mmin","vam","modi")
 }
 
-impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim", comp="rw_dist",donor_limit=Inf,optimal_donor="no",
+impute.NN_HD<-function(DATA=NULL,distance="man",weights="range",attributes="sim", comp="rw_dist",donor_limit=Inf,optimal_donor="no",
                 list_donors_recipients = NULL)
 {
-   original_data<-data
-   n<-dim(data)[1]
-   m<-dim(data)[2]
+   original_data<-DATA
+   n<-dim(DATA)[1]
+   m<-dim(DATA)[2]
    
    #test what type of weights and create appropriate vector
    if(length(weights)==1)
@@ -69,14 +72,14 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
          else
          {
             if(weights=="range")
-            {weights <-1/apply(apply(data,MARGIN=2,range,finite=TRUE),MARGIN=2,diff)}
+            {weights <-1/apply(apply(DATA,MARGIN=2,range,finite=TRUE),MARGIN=2,diff)}
             else
             {
                if(weights =="sd")
-               {weights <-1/apply(data,MARGIN=2,sd,na.rm=TRUE)}
+               {weights <-1/apply(DATA,MARGIN=2,sd,na.rm=TRUE)}
                else
                {#must be == "var"
-                  weights <-1/apply(data,MARGIN=2,var,na.rm=TRUE)
+                  weights <-1/apply(DATA,MARGIN=2,var,na.rm=TRUE)
                }
             }
          }
@@ -87,7 +90,7 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
          if(is.numeric(weights))
          {weights<-rep(weights,m)}
          else
-         {stop(paste("uniplemented weights used: ",weights))}
+         {stop(paste("unimplemented weights used: ",weights))}
       }
    }
    else
@@ -101,7 +104,7 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
             {stop("unimplemented feature, option 3 for weights\n")}
             else
             {
-               stop("inproper usage of option 3 for weights, number of weights must equal number of attributes\n")
+               stop("improper usage of option 3 for weights, number of weights must equal number of attributes\n")
             }
             weights<-rep(1,m)
          }
@@ -112,7 +115,7 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
             {stop("unimplemented feature, option 5 for weights\n")}
             else
             {
-               stop("inproper usage of option 5 for weights, number of weights must equal number of attributes\n")
+               stop("improper usage of option 5 for weights, number of weights must equal number of attributes\n")
             }
             weights<-rep(1,m)
          }
@@ -127,9 +130,9 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
       #if no distance matrix is specified, then donors and recipients need to be found
       if(is.null(list_donors_recipients))
       {
-         MV_Indicator_M<-matrix(.C("c_test_NA",data=as.double(data),
-                                   min=as.integer(ncol(data)),nin=as.integer(nrow(data)),
-                                   NAOK=TRUE)$data,nrow=nrow(data))
+         MV_Indicator_M<-matrix(.C("c_test_NA",data=as.double(DATA),
+                                   min=as.integer(ncol(DATA)),nin=as.integer(nrow(DATA)),
+                                   NAOK=TRUE)$data,nrow=nrow(DATA))
          NA_counts <- rowSums(MV_Indicator_M)
          if(attributes == "sim")
          {
@@ -161,7 +164,7 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
          #reweight data
          if(!all(weights==1))
          {
-            data<- reweight.data(data,weights,distance)
+           DATA<- reweight.data(DATA,weights,distance)
          }
          
          #print(list_donors_recipients)
@@ -170,7 +173,7 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
             #if(comp=="rw_dist")
             #{#no action needed}
             if(comp=="mean")
-            {data<-impute.mean(data)}
+            {DATA<-impute.mean(DATA)}
             if(comp=="rseq")
             {stop("unimplemented feature, comp = \"rseq\" \n")}
             if(comp=="rsim")
@@ -179,9 +182,9 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
          
          out <- .C("c_dist_hot_deck",
                    distances = as.double(rep(0,length(list_donors_recipients$donors)*length(list_donors_recipients$recipients))),
-                   data = as.double(data),
-                   n = as.integer(dim(data)[1]),
-                   m = as.integer(dim(data)[2]),
+                   data = as.double(DATA),
+                   n = as.integer(dim(DATA)[1]),
+                   m = as.integer(dim(DATA)[2]),
                    donors = as.integer(list_donors_recipients$donors-1),
                    recipients = as.integer(list_donors_recipients$recipients-1),
                    n_donors = as.integer(length(list_donors_recipients$donors)),
@@ -209,7 +212,7 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
             donor_limit<-ceiling(donor_limit)
          }
       }else{
-         stop(paste("inproper usage of donor_limit: ",donor_limit))}
+         stop(paste("improper usage of donor_limit: ",donor_limit))}
       if(is.infinite(donor_limit))
       {donor_limit<-length(list_donors_recipients$recipients)}
       
@@ -217,7 +220,7 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
       {stop(paste("Donor limit set too stringent. Number of donors * donor-limit must be >= Number of recipients.\n Compare:",
                   donor_limit*length(list_donors_recipients$donors),"vs.",length(list_donors_recipients$recipients)))
       }
-      optimal_donor = match.arg(arg=optimal_donor,choices=c("no", "rand", "mmin","modifvam","vam", "modi"),several.ok=FALSE)
+      optimal_donor = match.arg(arg=optimal_donor,choices=c("no", "rand", "mmin","modifvam","vam", "odd", "modi"),several.ok=FALSE)
       ##match.donor_recipient
       if(optimal_donor == "no")
       {
@@ -277,6 +280,12 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
          donor_index<-apply(distribution,2,which.max)
          list_recip_donor<-cbind(recipient=list_donors_recipients$recipients,donor=list_donors_recipients$donors[donor_index])
          
+      }
+      if(optimal_donor == "odd")
+      {
+         list_recip_donor<-match.d_r_odd(distance = distance,recipients=list_donors_recipients$recipients,
+                                         donors=list_donors_recipients$donors,donor_limit=rep(donor_limit,dim(distance)[1]))
+         
       }
       if(optimal_donor == "vam")
       {
@@ -288,44 +297,44 @@ impute.NN_HD<-function(data=NULL,distance="man",weights="range",attributes="sim"
       {stop("unimplemented feature, optimal_donor = \"modi\" \n")}
       
       #IMPUTE!
-         data <- .C("c_impute_NN_HD",
+         DATA <- .C("c_impute_NN_HD",
                    data = as.double(original_data),
-                   n = as.integer(dim(data)[1]),
-                   m = as.integer(dim(data)[2]),
+                   n = as.integer(dim(DATA)[1]),
+                   m = as.integer(dim(DATA)[2]),
                    recipients = as.integer(list_recip_donor[,1]-1),
                    n_recipients = as.integer(dim(list_recip_donor)[1]),
                    donors = as.integer(list_recip_donor[,2]-1),
                    NAOK=TRUE)$data
-         data<-matrix(data,nrow=n,ncol=m)
+         DATA<-matrix(DATA,nrow=n,ncol=m)
 }
    else
    {stop("Unimplemented way of handling attributes used: \"seq \"")}
    
-   return(data)
+   return(DATA)
 }
 
-impute.mean <-function(data=NULL)
+impute.mean <-function(DATA=NULL)
 {
    out <- .C("c_impute_mean",
-             data = as.double(data),
-             n = as.integer(dim(data)[1]),
-             m = as.integer(dim(data)[2]),
+             data = as.double(DATA),
+             n = as.integer(dim(DATA)[1]),
+             m = as.integer(dim(DATA)[2]),
              NAOK=TRUE
              )
-   return(matrix(out[[1]],ncol=dim(data)[2]))
+   return(matrix(out[[1]],ncol=dim(DATA)[2]))
 }
 
-reweight.data<-function(data=NULL,weights=NULL,minkovski_factor=1)
+reweight.data<-function(DATA=NULL,weights=NULL,minkovski_factor=1)
 {
    weights <- (weights)^(1/minkovski_factor)
    out <- .C("reweight_data",
-             data = as.double(data),
-             n = as.integer(dim(data)[1]),
-             m = as.integer(dim(data)[2]),
+             data = as.double(DATA),
+             n = as.integer(dim(DATA)[1]),
+             m = as.integer(dim(DATA)[2]),
              weights = as.double(weights),
              NAOK=TRUE
              )
-   return(matrix(out[[1]],ncol=dim(data)[2]))
+   return(matrix(out[[1]],ncol=dim(DATA)[2]))
 }
 
 # matrix, vector, vector, vector
@@ -382,4 +391,41 @@ match.d_r_vam<-function(distance = NULL,recipients=NULL,donors=NULL,donor_limit=
    }
    donor_index<-apply(distribution,2,which.max)
    return(cbind(recipient=recipients,donor=donors[donor_index]))
-}
\ No newline at end of file
+}
+
+match.d_r_odd<-function(distance = NULL,recipients=NULL,donors=NULL,donor_limit=NULL)
+{
+   #Get the number of donors and recipients
+   nd<-nrow(distance)
+   nr<-ncol(distance)
+   #setup the coefficient matrix A
+   A<-matrix(0,nrow=nr+nd,ncol=nr*nd)
+   #recipient requirements; colsums of distribution matrix must = 1
+   for(i in 1:nr)
+   {A[i,]<-c(rep(0,(i-1)*nd),rep(1,nd),rep(0,(nr-i)*nd))}
+   #donor limit; rowsums of distribution matrix must <= donor_limit
+   for(i in (nr+1):(nr+nd))
+   {A[i,]<-c(rep(0,(i-nr-1)),rep(c(1,rep(0,nd-1)),nr-1),1,rep(0,nd - (i-nr)))}
+   
+   #coefficients of objective function correspond to the distances
+   obj=as.vector(distance);
+
+   #recipient requirements and donor limit as explained for A
+   dir=c(rep("==",nr), rep("<=",nd));
+   
+   #recipient requirements and donor limit as explained for A
+   rhs=c(rep(1,nr),donor_limit);
+
+   #all variables are of integer type (maybe sometimes binary but never continuous)
+   types = rep("I",nd*nr);
+   
+   #pass LP that was setup above to through the interface to glpk
+   res<-Rglpk_solve_LP(obj,mat=A, dir, rhs, types, max = FALSE)
+   
+   #make a distribution matrix back out of the values of the base variables
+   distribution<-matrix(res$solution,nrow=nd)
+   
+   #get the corresponding donors
+   donor_index<-apply(distribution,2,which.max)
+   return(cbind(recipient=recipients,donor=donors[donor_index]))
+}
